//-----------------------------------------------------------------------------
//
//	Musepack Demuxer
//
//	Author : Igor Janos
//
//-----------------------------------------------------------------------------
#include "stdafx.h"



//-----------------------------------------------------------------------------
//
//	CMPCFile class
//
//-----------------------------------------------------------------------------

CMPCFile::CMPCFile() :
	duration_10mhz(0),
	reader(NULL),
	packet(NULL)
{
	key = 0;
	packet_max_size = 512 * 1024;			// well. hard to say. but I think this should be enough
	packet_size = 0;
	packet = (uint8*)malloc(packet_max_size);

	ASSERT(packet);
}

CMPCFile::~CMPCFile()
{
	if (packet) {
		free(packet);
		packet = NULL;
	}
}

// I/O for MPC file
int CMPCFile::Open(CMPCReader *reader)
{
	HRESULT		hr;
	int			ret;

	// keep a local copy of the reader
	this->reader = reader;


	/*
		According to stream specification the first 4 bytes should be 'MPCK'
	*/
	uint32	magick;
	reader->Seek(0);
	ret = reader->GetMagick(magick);			if (ret < 0) return ret;
	if (magick != 0x4d50434b) {
		// not a Musepack file
		return -1;
	}

	// now loop through a few packets
	do {
		int ret = ReadNextPacket();
		if (ret < 0) return -1;			// end of file reached before initialization is done ?

		switch (key) {
		case MPC_KEY('S','H'):	ret = ReadStreamHeader(); break;
		case MPC_KEY('R','G'):  ret = ReadReplaygain(); break;

		// TODO: Seek Table

		// audio packet - initialization is over
		case MPC_KEY('A','P'):	return 0;
		}

		// if parsing failed we quit
		if (ret < 0) return ret;
	} while (1);

	return 0;
}

int CMPCFile::ReadNextPacket()
{
	uint16	key_val;
	int64	size_val;
	int32	size_len;
	int		ret;

	ret = reader->GetKey(key_val);					if (ret < 0) return ret;
	ret = reader->GetSizeElm(size_val, size_len);	if (ret < 0) return ret;

	// if the key is not valid, quit
	if (!reader->KeyValid(key_val)) return -1;
	key = key_val;

	// size_val is the total size including key, variable size field and payload
	packet_size = size_val - 2 - size_len;

	// !!!!
	ASSERT(packet_size <= packet_max_size);

	ret = reader->Read(packet, packet_size);		if (ret < 0) return ret;
	return 0;
}

// parsing packets

int CMPCFile::ReadReplaygain()
{
	Bitstream	b(packet);
	b.NeedBits();

	int version = b.UGetBits(8);
	if (version != 1) return 0;		// unsupported RG version. not critical to us...

	int16	val;
	b.NeedBits();	val = b.SGetBits(16);	gain_title_db = val / 256.0;
	b.NeedBits();	val = b.SGetBits(16);	gain_title_peak_db = val / 256.0;
	b.NeedBits();	val = b.SGetBits(16);	gain_album_db = val / 256.0;
	b.NeedBits();	val = b.SGetBits(16);	gain_album_peak_db = val / 256.0;

	return 0;
}

int CMPCFile::ReadStreamHeader()
{
	Bitstream	b(packet);
	b.NeedBits();

	// let's do some reading
	uint32		crc;
	crc = b.UGetBits(16);	b.NeedBits();
	crc <<= 16;
	crc |= b.UGetBits(16);

	// TODO: CRC checking.
	b.NeedBits();
	stream_version = b.UGetBits(8);

	switch (stream_version) {
	case 7:
		{
			//-----------------------------------------------------------------
			//	Stream Version 7
			//-----------------------------------------------------------------

			// since we don't do any actual decoding here, we can only look 
			// for the fields of our interest

			// TODO: later. now we want SV8

		}
		break;
	case 8:
		{
			//-----------------------------------------------------------------
			//	Stream Version 8
			//-----------------------------------------------------------------
			int64	samples = b.GetMpcSize();
			int64	silence = b.GetMpcSize();

			b.NeedBits();
			uint8	freq = b.UGetBits(3);
			switch (freq) {
			case 0:	sample_rate = 44100; break;
			case 1:	sample_rate = 48000; break;
			case 2:	sample_rate = 37800; break;
			case 3: sample_rate = 32000; break;
			default:
				// invalid value
				return -1;				
			}

			// let's calculate duration
			duration_10mhz = (samples * 10000000) / sample_rate;

			b.DumpBits(5);					// bands
			channels = b.UGetBits(4)+1;		// channels

			b.NeedBits();
			b.DumpBits(1);					// mid side
			int exp = b.UGetBits(3);
			audio_block_frames = 1 << (2*exp);
		}
		break;
	default:
		// stream version not supported
		return -2;
	}

	// everything is okay
	return 0;
}


